#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>


enum exec_result {
	PARSE_OK,
	ERR_EXPECTED_INT,
	ERR_EXPECTED_REF,
	ERR_EXPECTED_QUOTE,
	ERR_STACK_EMPTY,
	ERR_INVALID_INT,
	ERR_INVALID_LITERAL,
	ERR_EOF
};

const char *exec_err(enum exec_result res) {
	switch(res) {
		case PARSE_OK:
			return "parsed successfully";
		case ERR_INVALID_LITERAL:
			return "invalid literal";
		case ERR_EXPECTED_INT:
			return "expected int on the stack";
		case ERR_EXPECTED_REF:
			return "expected reference on the stack";
		case ERR_EXPECTED_QUOTE:
			return "expected quote on the stack";
		case ERR_STACK_EMPTY:
			return "tried to pop empty stack";
		case ERR_INVALID_INT:
			return "invalid int";
		case ERR_EOF:
			return "end of file";
	}
	return "";
}

struct string {
	char *data;
	size_t size;
	size_t index;
};

struct string load_file(char *filename) {
	struct string str = {0};
	FILE *fp = fopen(filename, "r");
	if (!fp) exit(1);
	fseek(fp, 0, SEEK_END);
	str.size = ftell(fp);
	fseek(fp , 0, SEEK_SET);
	str.data = calloc(str.size + 1, sizeof(char));
	fread(str.data, 1, str.size, fp);
	fclose(fp);

	return str;
}

struct value {
	enum {
		VALUE_INT,
		VALUE_REF,
		VALUE_QUOTE
	} type;
	union {
		int32_t i;
		int ref_index;
		char *quote;
	};
};

struct program {
	struct value **stack;
	size_t nvals, maxvals;
	struct value *vars[26];
};

struct program *program_init() {
	struct program *p = malloc(sizeof(*p));
	p->maxvals = 1024;
	p->stack = calloc(p->maxvals, sizeof(*p->stack));
	p->nvals = 0;
	return p;
}

struct value *pop(struct program *stack) {
	if (stack->nvals == 0) {
		return NULL;
	}
	stack->nvals--;
	struct value *val = stack->stack[stack->nvals];
	stack->stack[stack->nvals] = NULL;

	return val;
}

void push(struct program *stack, struct value *val) {
	if (stack->nvals + 1 >= stack->maxvals) {
		stack->maxvals *= 2;
		stack->stack = realloc(stack->stack, stack->maxvals * sizeof(*stack->stack));
	}

	stack->stack[stack->nvals++] = val;
}

struct value *make_integer(int32_t val) {
	struct value *sval = malloc(sizeof(*sval));
	sval->type = VALUE_INT;
	sval->i = val;
	return sval;
}

struct value *make_reference(char val) {
	struct value *sval = malloc(sizeof(*sval));
	sval->type = VALUE_REF;
	sval->ref_index = val - 'a';
	return sval;
}

struct value *make_quote(char *val) {
	struct value *sval = malloc(sizeof(*sval));
	sval->type = VALUE_QUOTE;
	sval->quote = val;
	return sval;
}

struct value *dup_value(struct value *src) {
	struct value *sval = malloc(sizeof(*sval));
	*sval = *src;
	if (src->type == VALUE_QUOTE) {
		sval->quote = strdup(src->quote);
	}
	return sval;
}

void free_value(struct value *val) {
	if (val->type == VALUE_QUOTE)
		free(val->quote);
	free(val);
}

enum exec_result execute_program(struct program *p, struct string *data) {
	enum exec_result ret = PARSE_OK;
	char *end = data->data + data->size;
	for (char *c = data->data; c < end; c++) {
		/* skip whitespaces */
		for (; c < end && (*c == '\t' || *c == '\n' || *c == ' '); c++);

		/* if we end at a whitespace it won't break */
		if (c == end || *c == '\0') break;

		switch (*c) {
			case '{':
				for (; c < end && *c != '}'; c++);
				break;
			case '\'':
				if (c + 1 >= end) {
					ret = ERR_EOF;
					goto end;
				}
				push(p, make_integer(*(++c)));
				break;
			case '"':
				for (++c; c < end && *c != '"'; c++)
					fputc(*c, stdout);
				break;
			case '[': {
				char *closing = ++c;
				int nest = 0;
				for (; closing < end; closing++) {
					if (*closing == '[') {
						nest++;
					} else if (*closing == ']' && nest-- == 0) {
						break;
					}
				};
				push(p, make_quote(strndup(c, closing - c)));
				c = closing;
				break;
			}
			case ';': {
				struct value *ref = pop(p);
				if (ref->type != VALUE_REF) {
					ret = ERR_EXPECTED_REF;
					free_value(ref);
					goto end;
				}
				push(p, dup_value(p->vars[ref->ref_index]));
				free_value(ref);
				break;
			}
			case ':': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *ref = pop(p);
				struct value *val = pop(p);
				if (ref->type != VALUE_REF) {
					ret = ERR_EXPECTED_REF;
					free_value(ref);
					free_value(val);
					goto end;
				}
				p->vars[ref->ref_index] = val;
				free_value(ref);
				break;
			}
			case '$':
				push(p, dup_value(p->stack[p->nvals - 1]));
				break;
			case '%':
				free_value(pop(p)); // todo memory leak
				break;
			case '\\': {
				if (p->nvals < 2)
					exit(1);
				struct value *tmp = p->stack[p->nvals - 1];
				p->stack[p->nvals - 1] = p->stack[p->nvals - 2];
				p->stack[p->nvals - 2] = tmp;
				break;
			}
			case '@': {
				if (p->nvals < 3)
					exit(1);
				struct value *tmp = p->stack[p->nvals - 3];
				p->stack[p->nvals - 3] = p->stack[p->nvals - 2];
				p->stack[p->nvals - 2] = p->stack[p->nvals - 1];
				p->stack[p->nvals - 1] = tmp;
				break;
			}
			case 'O': {
				struct value *ind = pop(p);
				if (ind->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(ind);
					goto end;
				}
				if (ind->i > p->nvals) {
					ret = ERR_STACK_EMPTY;
					free_value(ind);
					goto end;
				}
				push(p, dup_value(p->stack[p->nvals - ind->i - 1]));
				free_value(ind);
				break;
			}
			case '+': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, make_integer(v1->i + v2->i));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '-': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, make_integer(v2->i - v1->i));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '*': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, make_integer(v1->i * v2->i));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '/': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, make_integer(v2->i / v1->i));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '_': {
				if (p->nvals < 1) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *val = pop(p);
				if (val->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(val);
					goto end;
				}
				push(p, make_integer(-val->i));
				free_value(val);
				break;
			}
			case '&': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, make_integer(v2->i & v1->i));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '|': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, make_integer(v2->i | v1->i));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '~': {
				if (p->nvals < 1) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				if (v1->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					goto end;
				}
				push(p, make_integer(~v1->i));
				free_value(v1);
				break;
			}
			case '=': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, v1->i == v2->i ? make_integer(-1) : make_integer(0));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '>': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_INT || v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				push(p, v2->i > v1->i ? make_integer(-1) : make_integer(0));
				free_value(v1);
				free_value(v2);
				break;
			}
			case '!': {
				if (p->nvals < 1) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				if (v1->type != VALUE_QUOTE) {
					ret = ERR_EXPECTED_QUOTE;
					free_value(v1);
					goto end;
				}
				struct string s = { .data = v1->quote, .size = strlen(v1->quote) };
				ret = execute_program(p, &s);
				free_value(v1);
				if (ret != PARSE_OK) {
					goto end;
				}
				break;
			}
			case '?': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_QUOTE) {
					ret = ERR_EXPECTED_QUOTE;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				if (v2->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				if (v2->i != 0) {
					struct string s = { .data = v1->quote, .size = strlen(v1->quote) };
					ret = execute_program(p, &s);
					if (ret != PARSE_OK) {
						free_value(v1);
						free_value(v2);
						goto end;
					}
				}
				free_value(v1);
				free_value(v2);
				break;
			}
			case '#': {
				if (p->nvals < 2) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *v1 = pop(p);
				struct value *v2 = pop(p);
				if (v1->type != VALUE_QUOTE || v2->type != VALUE_QUOTE) {
					ret = ERR_EXPECTED_QUOTE;
					free_value(v1);
					free_value(v2);
					goto end;
				}
				struct string body_quote = { .data = v1->quote, .size = strlen(v1->quote) };
				struct string cond_quote = { .data = v2->quote, .size = strlen(v2->quote) };
				ret = execute_program(p, &cond_quote);
				if (ret != PARSE_OK) {
					free_value(v1);
					free_value(v2);
					goto end;
				}
				struct value *cond = pop(p);
				if (cond->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(v1);
					free_value(v2);
					free_value(cond);
					goto end;
				}
				while (cond->i != 0) {
					free_value(cond);
					ret = execute_program(p, &body_quote);
					if (ret != PARSE_OK) {
						free_value(v1);
						free_value(v2);
						free_value(cond);
						goto end;
					}
					ret = execute_program(p, &cond_quote);
					if (ret != PARSE_OK) {
						free_value(v1);
						free_value(v2);
						free_value(cond);
						goto end;
					}
					cond = pop(p);
				}
				free_value(v1);
				free_value(v2);
				free_value(cond);
				break;

			}
			case '^': {
				char in = fgetc(stdin);
				if (in == EOF) {
					push(p, make_integer(-1));
				} else {
					push(p, make_integer(in));
				}
				break;
			}
			case ',': {
				if (p->nvals < 1) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *val = pop(p);
				if (val->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(val);
					goto end;
				}
				if (val->i > 255) {
					ret = ERR_INVALID_INT;
					free_value(val);
					goto end;
				}
				fputc(val->i, stdout);
				free_value(val);
				break;
			}
			case '.': {
				if (p->nvals < 1) {
					ret = ERR_STACK_EMPTY;
					goto end;
				}
				struct value *val = pop(p);
				if (val->type != VALUE_INT) {
					ret = ERR_EXPECTED_INT;
					free_value(val);
					goto end;
				}
				fprintf(stdout, "%d", val->i);
				free_value(val);
				break;
			}
			case 'B':
				break;
			case '`':
				break;
			default: {
				if (isalpha(*c)) {
					push(p, make_reference(*c));
				} else if (isdigit(*c)) {
					char *end;
					int32_t v = strtol(c, &end, 10);
					push(p, make_integer(v));
					c = end - 1;
				} else {
					ret = ERR_INVALID_LITERAL;
					goto end;
				}
				break;
			}
		}
	}

end:
	return ret;
}

int main(int argc, char **argv) {
	if (argc != 2)
		exit(1);

	struct string program = load_file(argv[1]);
	struct program *p = program_init();

	enum exec_result ret = execute_program(p, &program);

	if (ret != PARSE_OK) {
		printf("failed to interpret: %s\n", exec_err(ret));
	}

	return ret;
}
